#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['treeboost_autograd']

package_data = \
{'': ['*']}

install_requires = \
['numpy', 'torch']

extras_require = \
{'dev': ['catboost',
         'xgboost',
         'lightgbm',
         'matplotlib',
         'pytest',
         'scikit_learn']}

setup(name='treeboost_autograd',
      version='0.1.1',
      description='treeboost_autograd',
      author='Tomer Ronen',
      author_email='tomer.ronen34@gmail.com',
      url='https://github.com/TomerRonen34/treeboost_autograd',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
     )
